"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var XhrInterceptor = /** @class */ (function () {
    function XhrInterceptor() {
        // https://developer.mozilla.org/zh-CN/docs/Web/API/XMLHttpRequest
        this.xhr = window.XMLHttpRequest.prototype;
        this.mockData = {};
        this.intercept();
    }
    XhrInterceptor.prototype.setMockData = function (data) {
        this.mockData = data;
        return this;
    };
    XhrInterceptor.prototype.addMockData = function (key, val) {
        this.mockData[key] = val;
        return this;
    };
    XhrInterceptor.prototype.intercept = function () {
        // intercept methods
        this.interceptOpen();
        this.interceptSend();
        this.interceptSetRequestHeader();
        // intercept getters
        this.interceptReadyState();
        this.interceptStatus();
        this.interceptStatusText();
        this.interceptResponseText();
        this.interceptResponse();
        return this;
    };
    XhrInterceptor.prototype.matchRequest = function (reqUrl, reqMethod) {
        for (var key in this.mockData) {
            try {
                var info = this.mockData[key];
                if (info.disable === 'yes') {
                    continue;
                }
                var method = ("" + info.method).toLowerCase();
                if (method !== 'any' && method !== ("" + reqMethod).toLowerCase()) {
                    continue;
                }
                if (Array.isArray(info.regexp) && info.regexp.length === 2
                    && new RegExp(info.regexp[0], info.regexp[1]).test(reqUrl)) {
                    return info;
                }
                if ((info.url instanceof RegExp) && info.url.test(reqUrl)) {
                    return info;
                }
                if (reqUrl.indexOf(info.url) !== -1) {
                    return info;
                }
            }
            catch (e) { }
        }
        return false;
    };
    XhrInterceptor.prototype.interceptOpen = function () {
        var me = this;
        var original = this.xhr.open;
        Object.defineProperty(this.xhr, 'open', {
            get: function () {
                var _this = this;
                return function (method, url, async, user, password) {
                    var match = me.matchRequest(url, method);
                    if (match) {
                        _this.isMockRequest = true;
                        _this.mockRequestInfo = match;
                        return;
                    }
                    return original.call(_this, method, url, async, user, password);
                };
            }
        });
        return this;
    };
    XhrInterceptor.prototype.interceptSend = function () {
        var me = this;
        var original = this.xhr.send;
        Object.defineProperty(this.xhr, 'send', {
            get: function () {
                var _this = this;
                return function (body) {
                    if (_this.isMockRequest) {
                        return _this.mockRequestInfo.delay && _this.mockRequestInfo.delay > 0
                            ? setTimeout(function () { return me.doCompleteCallbacks(_this); }, +_this.mockRequestInfo.delay)
                            : me.doCompleteCallbacks(_this);
                    }
                    return original.call(_this, body);
                };
            }
        });
        return this;
    };
    XhrInterceptor.prototype.doCompleteCallbacks = function (xhrInstance) {
        xhrInstance.onreadystatechange();
        xhrInstance.onload && xhrInstance.onload();
        xhrInstance.onloadend && xhrInstance.onloadend();
    };
    XhrInterceptor.prototype.interceptSetRequestHeader = function () {
        var original = this.xhr.setRequestHeader;
        Object.defineProperty(this.xhr, 'setRequestHeader', {
            get: function () {
                var _this = this;
                return function (header, value) {
                    if (_this.isMockRequest) {
                        return;
                    }
                    return original.call(_this, header, value);
                };
            }
        });
        return this;
    };
    XhrInterceptor.prototype.getGetter = function (key) {
        var descriptor = Object.getOwnPropertyDescriptor(this.xhr, key);
        if (descriptor) {
            return descriptor.get;
        }
        // when XMLHttpRequest is not a standard implement.
        return this.xhr[key];
    };
    XhrInterceptor.prototype.interceptReadyState = function () {
        var original = this.getGetter('readyState');
        Object.defineProperty(this.xhr, 'readyState', {
            get: function () {
                if (this.isMockRequest) {
                    return 4;
                }
                return typeof original === 'function' ? original.call(this) : original;
            }
        });
        return this;
    };
    XhrInterceptor.prototype.interceptStatus = function () {
        var original = this.getGetter('status');
        Object.defineProperty(this.xhr, 'status', {
            get: function () {
                if (this.isMockRequest) {
                    return 200;
                }
                return typeof original === 'function' ? original.call(this) : original;
            }
        });
        return this;
    };
    XhrInterceptor.prototype.interceptStatusText = function () {
        var original = this.getGetter('statusText');
        Object.defineProperty(this.xhr, 'statusText', {
            get: function () {
                if (this.isMockRequest) {
                    return 'OK';
                }
                return typeof original === 'function' ? original.call(this) : original;
            }
        });
        return this;
    };
    XhrInterceptor.prototype.interceptResponseText = function () {
        var original = this.getGetter('responseText');
        Object.defineProperty(this.xhr, 'responseText', {
            get: function () {
                if (this.isMockRequest) {
                    var data = this.mockRequestInfo.data;
                    return typeof data === 'string' ? data : JSON.stringify(data);
                }
                return typeof original === 'function' ? original.call(this) : original;
            }
        });
        return this;
    };
    XhrInterceptor.prototype.interceptResponse = function () {
        var original = this.getGetter('response');
        Object.defineProperty(this.xhr, 'response', {
            get: function () {
                if (this.isMockRequest) {
                    return this.responseType !== '' ? this.mockRequestInfo.data : this.responseText;
                }
                return typeof original === 'function' ? original.call(this) : original;
            }
        });
        return this;
    };
    return XhrInterceptor;
}());
exports.default = XhrInterceptor;
//# sourceMappingURL=XhrInterceptor.js.map